@@ -45,7 +45,7 @@ class ProgressBar():
         if tblink_max >= 1:
             tblink_max = .9
             print("'tblink_max' was set to .9 seconds beacuse of exceeding maximum limit!")
-        self.desc = desc[:front_space] + ' ' + '-' * (front_space - len(desc) - 2) + ' '
+        self.desc = desc[:front_space] + ' ' + '-' * (front_space - len(desc)) * int(front_space > len(desc))+ ' '
         self.barend = ' '*15
         self.tblmax = tblink_max
         self.tblmin = tblink_min
@@ -92,6 +92,8 @@ class ProgressBar():
         if self.tcntst:
             d = self.leftime.days
             s = int(self.leftime.seconds - self.tstack)
+            if d < 0:
+                d, s = 0, 0
             self.prleftime = '%s'%datetime.timedelta(days = d, seconds = s)
     
     def set_barelap(self):
@@ -131,6 +133,7 @@ class ProgressBar():
     def __exit__(self, exception_type, exception_value, traceback):
         self.flblink = False
         time.sleep(self.tblmax)
+        self.thread.join()
         self.barend = ' Complete!' + ' '*15
         self.barprint('\n')
 
@@ -292,7 +295,7 @@ class Filters():
 # savepath : str, path to save the figure, default is ''
 # fname : str, name of the figure to save in the savepath, default is fig_title
 
-def get_spectrogram(ts_sr, fs = None, win = ('tukey', 0.25), nperseg = None, noverlap = None, mode = 'psd', figsize = None, vis_frac = 1, ret_sxx = False, show = True, save = False, saveapth = '', fname = ''):
+def get_spectrogram(ts_sr, fs = None, win = ('tukey', 0.25), nperseg = None, noverlap = None, mode = 'psd', figsize = None, vis_frac = 1, ret_sxx = False, show = True, save = False, savepath = '', fname = ''):
    
     # default values
     if fs == None: fs = 1
@@ -405,6 +408,30 @@ def get_edges_from_ts(sr, th_method = 'median', th_factor = .5, th = None,  del_
 
 
 
+# ############## EACH_ROW_MAX ###############
+# this function gets the maximum values and corresponding columns for each row of a matrix
+# data : list of lists/numpy ndarray or pandas dataframe, matrix data from where max values will be calculated
+# returns same data with two new columns with max values and corresponding column names
+def each_row_max(data):
+    if isinstance(data, np.ndarray) or isinstance(data, list): data = pd.DataFrame(data)
+    elif isinstance(data, pd.DataFrame): pass
+    else: msd.InputVariableError('input data type must be list of lists or pandas dataframe or numpy ndarray!')
+    if data.isnull().all().all():
+        data['max_val'] = np.nan
+        data['max_col'] = np.nan
+    else:
+        col = data.columns
+        row = data.index
+        data = data.values
+        max_idx = np.nanargmax(data, axis = 1)
+        max_val = [data[i, max_idx[i]] for i in range(data.shape[0])]
+        max_col = col[max_idx]
+        data = pd.DataFrame(data, index = row, columns = col)
+        data['max_val'] = max_val
+        data['max_col'] = max_col
+    return data
+
+
 
 # moving slope is a function which calculates the slope inside a window for a variable
 # df : pandas DataFrame or Series, contains time series columns
@@ -485,7 +512,8 @@ def standardize(data, zero_std = 1):
 # #####  NORMALIZE  #############
 # data : pandas series, dataframe, list or numpy ndarray, input data to be standardized
 # zero_range : float, value used to replace range values in case range is 0 for any column
-def normalize(data, zero_range = 1):
+# method : {'zero_mean', 'min_max_0_1', 'min_max_-1_1'}
+def normalize(data, zero_range = 1, method = 'min_max_0_1'):
     dtype = 0
     if isinstance(data, pd.Series):
         data = data.to_Frame()
@@ -497,7 +525,9 @@ def normalize(data, zero_range = 1):
     else: data_type_error('Provided data is inappropriate! ')
     data_range = data.max() - data.min()
     data_range[data_range == 0] = zero_range
-    data = (data - data.mean()) / data_range
+    if method == 'min_max_0_1': data = (data - data.min()) / data_range
+    elif method == 'min_max_-1_1': data = (data - data.min()) / data_range * 2 - 1
+    elif method == 'zero_mean': data = (data - data.mean()) / data_range
     if dtype == 0: return data
     elif dtype == 1: return data[data.columns[0]]
     else : return data.values
@@ -557,10 +587,17 @@ def name_separation(string, maxlen):
 # spalpha : list of constants indicating alpha values for each of the span in spans
 # ylims : list of lists, indicating y axis limits in case we want to keep the limit same for all subplots
 # name_thres : maximum allowed characters in one line for plot labels
+# fig_x: float, horizontal length of the figure, default is 30
+# fig_y: float, vertical length of each row of plot, default is 3
+# marker: str, marker for time series plots, default is None
+# xlabel : str, label name for x axis for each row, default is 'Time'
+# ylabel : str, label name for y axis for each row, default is 'Data value'
 
 def input_variable_error(msg):
     raise msdExceptions.InputVariableError(msg)
-def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans = [], lines = [], linestyle = [], linewidth = [], fig_title = '', show = True, save = False, savepath = '', fname = '', spine_dist = .035, spalpha = [], ylims = [], name_thres = 50):
+def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans = [], lines = [], linestyle = [], linewidth = [],
+                     fig_title = '', show = True, save = False, savepath = '', fname = '', spine_dist = .035, spalpha = [],
+                     ylims = [], name_thres = 50, fig_x = 30, fig_y = 3, marker = None, xlabel = 'Time', ylabel = 'Data value'):
    
     totsame = len(same_srs)
     totdif = len(srs)
@@ -593,11 +630,15 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
     else: input_variable_error('Invalid line width! linewidth should be int/float or list of int/float with equal length of lines')
    
     nrows = segs.shape[0]
-    colors = ['darkcyan', 'coral', 'darkslateblue', 'limegreen', 'crimson', 'purple', 'blue', 'khaki', 'chocolate']
-    spcolors = ['darkcyan', 'coral', 'purple', 'red', 'khaki', 'gray']
-    lcolors = ['crimson', 'magenta', 'darkolivegreen', 'palevioletred', 'indigo']
+    colors = ['darkcyan', 'coral', 'darkslateblue', 'limegreen', 'crimson', 'purple', 'blue', 'khaki', 'chocolate', 'forestgreen']
+    spcolors = ['darkcyan', 'coral', 'purple', 'red', 'khaki', 'gray', 'darkslateblue', 'limegreen', 'red', 'blue']
+    lcolors = ['crimson', 'magenta', 'darkolivegreen', 'palevioletred', 'indigo', 'chokolate', 'blue', 'forestgreen', 'grey', 'magenta']
+    if totsame + totdif > len(colors): colors = get_color_from_cmap(totsame + totdif, 'rainbow')
+    if totsp > len(spcolors): colors = get_color_from_cmap(totsp, 'rainbow')
+    if totline > len(lcolors): colors = get_color_from_cmap(totline, 'rainbow')
     if spalpha == []: spalpha = [.3 for _ in range(totsp)]
-    fig, ax = plt.subplots(figsize = (30, 3 * nrows), nrows = nrows)
+    stamp_fl = True if isinstance(same_srs[0].index[0], pd.Timestamp) else False
+    fig, ax = plt.subplots(figsize = (fig_x, fig_y * nrows), nrows = nrows)
     if fig_title == '': fig_title = 'Time Series Visualization'
     fig.suptitle(fig_title, y = 1.03, fontsize = 20, fontweight = 'bold')
     if nrows == 1: ax = [ax]
@@ -618,7 +659,7 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
         # time series plots
         for j in range(totdif):
             tx = ax[i].twinx()
-            l, = tx.plot(srs[j].loc[st : ed], color = colors[totsame + j], zorder = 30 + j * 4)
+            l, = tx.plot(srs[j].loc[st : ed], color = colors[totsame + j], zorder = 30 + j * 4, marker = marker)
             lg[0].append(l)
             lg[1].append(srs[j].name)
             tx.spines['right'].set_position(('axes', 1 + j * spine_dist))
@@ -629,7 +670,7 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
             if len(ylims) > 1:
                 if len(ylims[1 + j]) == 2: tx.set_ylim(ylims[1 + j][0], ylims[1 + j][1])
         for j in range(totsame):
-            l, = ax[i].plot(same_srs[j].loc[st : ed], color = colors[j], zorder = j * 4, lw = same_srs_width[j])
+            l, = ax[i].plot(same_srs[j].loc[st : ed], color = colors[j], zorder = j * 4, lw = same_srs_width[j], marker = marker)
             lg[0].append(l)
             lg[1].append(same_srs[j].name)
         
@@ -660,9 +701,9 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
                 lg[1].append(spans[j].columns.name)
        
         # finishing axis arrangements
-        ax[i].set_xlabel('Time')
-        ax[i].set_ylabel('Data Value' if totsame > 1 else name_separation(same_srs[0].name, name_thres))
-        ax[i].set_title('from %s to %s'%(st.strftime('%Y-%m-%d %H-%M-%S'), ed.strftime('%Y-%m-%d %H-%M-%S')), loc = 'right')
+        ax[i].set_xlabel(xlabel)
+        ax[i].set_ylabel(ylabel if totsame > 1 else name_separation(same_srs[0].name, name_thres))
+        if stamp_fl: ax[i].set_title('from %s to %s'%(st.strftime('%Y-%m-%d %H-%M-%S'), ed.strftime('%Y-%m-%d %H-%M-%S')), loc = 'right')
         ax[i].legend(lg[0], lg[1], loc = 3, bbox_to_anchor = (0, .98), ncol = len(lg[0]), fontsize = 10)
         ax[i].set_xlim(st, ed)
         ax[i].grid(True)
@@ -682,10 +723,12 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
 
 # data : pandas DataFrame, data to be plotted as heatmap
 # stdz : bool, whether to standardize the data or not, default is False
-# keep_only : {'both', 'up', 'down'}, which side of the heatmap matrix to plot, necessary for correlation heatmap plot, default is 'both'
+# keep : {'both', 'up', 'down'}, which side of the heatmap matrix to plot, necessary for correlation heatmap plot, default is 'both'
 # rem_diag : bool, whether to remove diagoanl if keep_only is not 'both', default is False
 # cmap : str, matplotlib colormap, default is 'gist_heat'
 # cbar : bool, show the colorbar with the heatmap or not
+# annotate : bool, whether to show the values or not
+# fmt : str, value format for printing if annotate is True
 # show : bool, show the heatmap or not
 # save : bool, save the figure or not
 # savepath : str, path for saving the figure
@@ -693,11 +736,11 @@ def plot_time_series(same_srs, srs = [], segs = None, same_srs_width = [], spans
 # fig_title : str, title of the heatmap, default is 'Heatmap of {data.columns.name}'
 # file_name : str, name of the image as will be saved in savepath, default is fig_title
 # lbfactor : float/int, factor used for scaling the plot labels
-def plot_heatmap(data, keep = 'both', rem_diag = False, cmap = 'gist_heat', cbar = True, stdz = False, 
+def plot_heatmap(data, keep = 'both', rem_diag = False, cmap = 'gist_heat', cbar = True, stdz = False, annotate = False, fmt = None,
                  show = True, save = False, savepath = '', figsize = (30, 10), fig_title = '', file_name = '',
-                 lbfactor = 1.5):
-    xlb = data.index.name
-    ylb = data.columns.name
+                 lbfactor = 1.5, xrot = 90):
+    ylb = data.index.name
+    xlb = data.columns.name
     if stdz:
         data_std = data.std()
         data_std[data_std == 0] = 1
@@ -712,11 +755,11 @@ def plot_heatmap(data, keep = 'both', rem_diag = False, cmap = 'gist_heat', cbar
         k = -1 if rem_diag else 0
         data = data.where(np.tril(np.ones(data.shape), k = k).astype(bool))
     fig, ax = plt.subplots(figsize = figsize)
-    sns.heatmap(data, ax = ax, linewidths = 0, cbar = cbar, cmap = cmap)
+    sns.heatmap(data, ax = ax, linewidths = 0, cbar = cbar, cmap = cmap, annot = annotate, fmt = fmt)
     ax.set_xlabel(xlb)
     ax.set_ylabel(ylb)
     ax.tick_params(axis = 'y', rotation = 0)
-    ax.tick_params(axis = 'x', rotation = 90)
+    ax.tick_params(axis = 'x', rotation = xrot)
     if fig_title == '': fig_title = 'Heatmap of %s'%data.columns.name if data.columns.name not in ['', None] else 'Heatmap'
     ax.set_title(fig_title)
     fig.tight_layout()
@@ -804,7 +847,7 @@ def data_gridplot(data, idf = [], idf_pref = '', idf_suff = '', diag = 'hist', b
             if j == 0: ax[i, j].set_ylabel(col[i])
             ax[i, j].grid(True)
     fig.tight_layout()
-    plt.show()
+    if show: plt.show()
     if save and savepath != '':
         if fname == '': fname = fig_title
         fig.savefig('%s/%s.jpg'%(savepath, fname), bbox_inches = 'tight')
@@ -829,6 +872,14 @@ def plot_diag(arr, bins, ax, diag, color, label):
 def get_color_from_cmap(n, cmap = 'jet', rng = [.05, .95]):
     return [get_cmap(cmap)(i) for i in np.linspace(rng[0], rng[1], n)]
 
+# returns many CSS4 named colors available in matplotlib
+def get_named_colors():
+    colors = ['darkcyan', 'crimson', 'coral', 'forestgreen', 'purple', 'magenta', 'khaki', 'maroon',
+              'darkslategray', 'brown', 'gray', 'darkorange', 'mediumseagreen', 'royalblue', 'midnightblue',
+              'turquoise', 'rosybrown', 'darkgoldenrod', 'olive', 'saddlebrown', 'darkseagreen', 'deeppink',
+              'aqua', 'lawngreen', 'tab:red', 'gold', 'tan', 'peru', 'violet', 'thistle', 'steelblue',
+              'darkkhaki', 'chocolate', 'mediumspringgreen', 'rebeccapurple', 'tomato']
+    return colors
 
 
 
@@ -847,8 +898,15 @@ def get_color_from_cmap(n, cmap = 'jet', rng = [.05, .95]):
 # fig_width: float, is the width of figure, default is 30
 # fig_height: float, is the height of figure, default is 5
 # rotation: float, rotational angle of x axis labels of the figure, default is 40
+# show : bool, whether to show the graph or not, default is True
+# save : bool, whether to save the figure or not, default is False
+# savepath : str, path where the data will be saved, default is None
+# fname : str, filename which will be used for saving the figure, default is None
+# fig_title : str, title of the figure, default is 'Feature Confidence Evaluation for categorical Features'
 # This function tries to calculate the importance of features by some statistical approaches
-def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num = False, is_all_cat = False, num_cols = [], cat_cols = [], cmap = 'gist_rainbow', fig_width = 30, fig_height = 5, rotation = 40):
+def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num = False, is_all_cat = False, num_cols = [], 
+                      cat_cols = [], cmap = 'gist_rainbow', fig_width = 30, fig_height = 5, rotation = 40, show = True, save = False, 
+                      savepath = None, fname = None, fig_title = ''):
     
     # separating numerical and categorical feature data 
     if is_all_num:
@@ -887,6 +945,7 @@ def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num =
     # loop definitions
     labels_list = [i for i in [num_labels, cat_labels] if i.shape[1] > 0]
     labels_tag = ['Numerical label', 'Categorical label']
+    fig_title = 'Feature Confidence Evaluation' if fig_title == '' else fig_title
     # to collect all values of evaluation
     num_lab_confids = {}
     # loop for numerical and categorical labels
@@ -896,10 +955,12 @@ def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num =
         nrows = labels.columns.shape[0]
         if not is_all_cat:
             fig_num, ax_num = plt.subplots(figsize = (fig_width, fig_height * nrows), nrows = nrows)
-            fig_num.suptitle('Feature Confidence Evaluation for Numerical Features', y = 1.06, fontsize = 20, fontweight = 'bold')
+            fig_tnum = fig_title + ' for Numerical Features'
+            fig_num.suptitle(fig_tnum, y = 1.06, fontsize = 20, fontweight = 'bold')
         if not is_all_num:
             fig_cat, ax_cat = plt.subplots(figsize = (fig_width, fig_height * nrows), nrows = nrows)
-            fig_cat.suptitle('Feature Confidence Evaluation for categorical Features', y = 1.06, fontsize = 20, fontweight = 'bold')
+            fig_tcat = fig_title + ' for Categorical Features'
+            fig_cat.suptitle(fig_tcat, y = 1.06, fontsize = 20, fontweight = 'bold')
         # special case for nrows = 1
         if nrows == 1:
             if not is_all_cat:
@@ -959,9 +1020,18 @@ def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num =
         
         if not is_all_cat:
             fig_num.tight_layout()
+            if save and savepath != None:
+                if fname == None: fname = fig_tnum
+                if savepath[-1] != '/': savepath += '/'
+                fig_num.savefig('%s%s.jpg'%(savepath, fname), bbox_inches = 'tight')
         if not is_all_num:
             fig_cat.tight_layout()
-        plt.show()
+            if save and savepath != None:
+                if fname == None: fname = fig_tcat
+                if savepath[-1] != '/': savepath += '/'
+                fig_cat.savefig('%s%s.jpg'%(savepath, fname), bbox_inches = 'tight')
+        if show: plt.show()
+        plt.close()
     return num_lab_confids
 
 
@@ -971,8 +1041,8 @@ def feature_evaluator(data, label_name, label_type_num, n_bin = 40, is_all_num =
 # ############ CLASS_RESULT  ##############
 # classification result with confusion matrix
 # ARGUMENTS
-# y --> True labels
-# pred --> Predictions
+# y --> True labels, a vector holding class indices like 0, 1, 2 .. etc.
+# pred --> Predictions, a vector containing prediction opt for class indices similar to True labels 
 # out_confus --> returns confusion matrix if set True
 # RETURNS
 # result --> DataFrame containing precision, recall and f1 scores for all labels
@@ -1014,7 +1084,7 @@ def class_result(y, pred, out_confus = False):
         else:
             f = 0
         f_msr.append(f)
-        acc.append('-')
+        acc.append(np.nan)
     result = pd.DataFrame([prec, recall, f_msr], columns = labels, index = ['precision', 'recall', 'f1_score'])
     avg = result.T.mean()
     avg.name = 'average'
@@ -1250,4 +1320,4 @@ class SplitDataset():
                     outdata[nm][k][s] = np.flipud(np.array(outdata[nm][k][s]))
                 if sec == 1 or nm == 'test':
                     outdata[nm][k] = outdata[nm][k]['section_1']
-        return outdata
\ No newline at end of file
+        return outdata
